// JNIArray.cpp

#ifndef __jniarray_cpp__
#define __jniarray_cpp__

#include <jni.h>
#include <typeinfo.h>
#include "WindowThread.h"
#include "JNIArray.h"


extern WindowThread threadInstance; // Thread instance from WindowsThread

template <typename T>
JNIArray<T>::JNIArray(jobject jObject, const char * name)
{
  if (jObject == 0)
  {
     SHOW_ERROR("Fatal error when creating JNIArray", "No object exposed");
  return;
  }
  _env = threadInstance.getEnv();
  _jObject = jObject;
  _pName = new char[strlen(name) + 1];
  strcpy(_pName, name);
  _type[0] = 0;
  _isAttached = false;

  if (strcmp(typeid(T).name(), "wchar_t") == 0)
    strcpy(_type, "[C");
  if (strcmp(typeid(T).name(), "bool") == 0)
    strcpy(_type, "[Z");
  if (strcmp(typeid(T).name(), "signed char") == 0)
    strcpy(_type, "[B");
  if (strcmp(typeid(T).name(), "short") == 0)
    strcpy(_type, "[S");
  if (strcmp(typeid(T).name(), "int") == 0)
    strcpy(_type, "[I");
  if (strcmp(typeid(T).name(), "__int64") == 0)
    strcpy(_type, "[J");
  if (strcmp(typeid(T).name(), "float") == 0)
    strcpy(_type, "[F");
  if (strcmp(typeid(T).name(), "double") == 0)
    strcpy(_type, "[D");

//  cout << "type: " << typeid(T).name() << endl;

  if (_type[0] == 0)
  {
    char msg[256];
    strcpy(msg, "Illegal type Got: ");
    strcat(msg, typeid(T).name());
    strcat(msg, "\nbut must be one of\n");
    strcat(msg, "bool, signed char, short, int, long, float, double, wchart_t\n");
    SHOW_ERROR("Fatal error when constructing JNIArray", msg);
    return;
  }
  _id = getFieldID();
}

template <typename T>
JNIArray<T>::~JNIArray()
{
  if (_pName)
  {
    delete [] _pName;
    release(); 
  }
}

template <typename T>
jfieldID JNIArray<T>::getFieldID()
{
   char msg[128];

//   cout << "looking for variable " << _pName << " with type: " << _type << endl;
   _id = _env->GetFieldID(_env->GetObjectClass(_jObject), _pName, _type);

   if (!_id)
   {
     strcpy(msg, "Arrayvariable ");
     strcat(msg, _pName);
     strcat(msg, " not found.");
  SHOW_ERROR("Fatal error in JNIArray::getFieldID()", msg);
     return 0;
   }
   return _id;
}

template <typename T>
T * JNIArray<T>::get()
{
   if (_isAttached)
   {
      SHOW_ERROR("Fatal error in JNIArray::get()",
           "Must release JNIArray before calling get() again");
      return NULL;
  }

  switch (_type[1])
  {
    case 'C':
    {
       _charAry = (jcharArray)_env->GetObjectField(_jObject, _id);
       _charArrayElements = _env->GetCharArrayElements(_charAry, &_isCopy);
       char * ary = (char *)_charArrayElements;
       _isAttached = true;
       return (T *)ary;
    }

    case 'Z':
    {
       _booleanAry = (jbooleanArray)_env->GetObjectField(_jObject, _id);
       _booleanArrayElements = _env->GetBooleanArrayElements(_booleanAry, &_isCopy);
       boolean * ary = (boolean *)_booleanArrayElements;
       _isAttached = true;
       return (T *)ary;
    }

    case 'B':
    {
       _byteAry = (jbyteArray)_env->GetObjectField(_jObject, _id);
       _byteArrayElements = _env->GetByteArrayElements(_byteAry, &_isCopy);
       BYTE * ary = (BYTE *)_byteArrayElements;
       _isAttached = true;
       return (T *)ary;
    }

    case 'S':
    {
       _shortAry = (jshortArray)_env->GetObjectField(_jObject, _id);
       _shortArrayElements = _env->GetShortArrayElements(_shortAry, &_isCopy);
       short * ary = (short *)_shortArrayElements;
       _isAttached = true;
       return (T *)ary;
    }

    case 'I':
    {
       _intAry = (jintArray)_env->GetObjectField(_jObject, _id);
       _intArrayElements = _env->GetIntArrayElements(_intAry, &_isCopy);
       int * ary = (int *)_intArrayElements;
       _isAttached = true;
       return (T *)ary;
    }

    case 'J':
    {
       _longAry = (jlongArray)_env->GetObjectField(_jObject, _id);
       _longArrayElements = _env->GetLongArrayElements(_longAry, &_isCopy);
       __int64 * ary = (__int64 *)_longArrayElements;
       _isAttached = true;
       return (T *)ary;
    }

    case 'F':
    {
       _floatAry = (jfloatArray)_env->GetObjectField(_jObject, _id);
       _floatArrayElements = _env->GetFloatArrayElements(_floatAry, &_isCopy);
       float * ary = (float *)_floatArrayElements;
       _isAttached = true;
       return (T *)ary;
    }

    case 'D':
    {
       _doubleAry = (jdoubleArray)_env->GetObjectField(_jObject, _id);
       _doubleArrayElements = _env->GetDoubleArrayElements(_doubleAry, &_isCopy);
       double * ary = (double *)_doubleArrayElements;
       _isAttached = true;
       return (T *)ary;
    }
  }
  SHOW_ERROR("Fatal error in JNIArray::get()", "Illegal type");
  return NULL;
}

template <typename T>
void JNIArray<T>::release()
{
   if (_isCopy == JNI_TRUE && _isAttached)
   {
      switch (_type[1])
      {
         case 'C':
           _env->ReleaseCharArrayElements(_charAry, _charArrayElements, 0);
           break;
         case 'Z':
           _env->ReleaseBooleanArrayElements(_booleanAry, _booleanArrayElements, 0);
           break;
         case 'B':
           _env->ReleaseByteArrayElements(_byteAry, _byteArrayElements, 0);
           break;
         case 'S':
           _env->ReleaseShortArrayElements(_shortAry, _shortArrayElements, 0);
           break;
         case 'I':
           _env->ReleaseIntArrayElements(_intAry, _intArrayElements, 0);
           break;
         case 'J':
           _env->ReleaseLongArrayElements(_longAry, _longArrayElements, 0);
           break;
         case 'F':
           _env->ReleaseFloatArrayElements(_floatAry, _floatArrayElements, 0);
           break;
         case 'D':
           _env->ReleaseDoubleArrayElements(_doubleAry, _doubleArrayElements, 0);
           break;
      }
   }
   _isAttached = false;   // Don't release twice
}


#endif

