// WindowHandler.h

#ifndef __WindowHandler_h__
#define __WindowHandler_h__

#include <windows.h>
#include <jni.h>

#define LPRESS      1
#define LRELEASE    2
#define LCLICK      4
#define LDCLICK     8
#define RPRESS     16
#define RRELEASE   32
#define RCLICK     64
#define RDCLICK   128
#define ENTER     256
#define LEAVE     512
#define MOVE     1024

#define ID_TIMER_ICON 8765

#include "WindowThread.h"
#include "JNIThread.h"
#include "JNIBuffer.h"

#ifndef WS_EX_LAYERED
#define WS_EX_LAYERED  0x00080000
#endif

#ifndef LWA_COLORKEY
#define LWA_COLORKEY  0x00000001
#endif
#ifndef LWA_ALPHA
#define LWA_ALPHA  0x00000002
#endif

#ifndef SHOW_ERROR
#define SHOW_ERROR(title, msg) \
   MessageBox(HWND_DESKTOP, msg, title, MB_ICONEXCLAMATION | MB_OK);
#endif

typedef DWORD (WINAPI *PSLWA)(HWND, DWORD, BYTE, DWORD);


class WindowHandler
{
friend DWORD WINAPI WindowThread::ThreadProc(LPVOID lpParameter);
friend DWORD WINAPI JNIThread::ThreadProc(LPVOID lpParameter);

public:
   WindowHandler(JNIEnv * env, jobject handlerObj);  // Ctor
   virtual ~WindowHandler();  // Dtor
   void enable(JNIEnv *env);
   void expose(jobject obj);
   void showWindow(bool isVisible, int percent, int red, int green, int blue, bool wait);
   void setWindowPosition(int xPos, int yPos);
   void setWindowSize(int width, int height);
   void disable();
   bool initDone() {return isInitDone;}
   virtual void createIconEventBuf(jintArray jary, jint size);
   virtual int readIconEventBuf(jintArray jary, int count);
   virtual void createBuf(jbooleanArray jary, jint size);
   virtual void createBuf(jcharArray jary, jint size);
   virtual void createBuf(jbyteArray jary, jint size);
   virtual void createBuf(jshortArray jary, jint size);
   virtual void createBuf(jintArray jary, jint size);
   virtual void createBuf(jlongArray jary, jint size);
   virtual void createBuf(jfloatArray jary, jint size);
   virtual void createBuf(jdoubleArray jary, jint size);
   virtual int readBuf(int count);
   virtual void writeBool(bool value);
   virtual void writeChar(wchar_t value);
   virtual void writeByte(signed char value);
   virtual void writeShort(short value);
   virtual void writeInt(int value);
   virtual void writeLong(__int64 value);
   virtual void writeFloat(float value);
   virtual void writeDouble(double value);
   virtual void flushBuf();
   virtual int countBuf();
   virtual bool isOverflow(bool reset);
   virtual int invoke(int tag);
   void setGrantObject(jobject jObject);
   jfieldID getFieldID(jobject jObject, const char * name, const char * type);
   jobject getObjectVariable(jobject jObject, const char * name, const char * classname);
   HWND getHwnd() {return _hwnd;}
   static WindowHandler * getThis(HWND hwnd);
   static jobject getObject(HWND hwnd);
   static jobject getAppObject(HWND hwnd);
   static void setInstance(HINSTANCE hInst) {_hInstance = hInst;}
   static HINSTANCE getInstance() {return _hInstance;}
   static WindowThread * getThread() {return _pThread;}
   void startJNIThread();
   void stopJNIThread(); 
   virtual void createTrayIcon();
   virtual void modifyTrayIcon();
   virtual void deleteTrayIcon();
   virtual void showBalloonTooltip(int timeout, int flag);
   virtual bool wantThreadMsg(UINT uMsg);
   virtual void evThreadMsg(HWND hWnd, UINT uMsg, WPARAM wParam,
                            LPARAM lParam);
   virtual void setImageLoaded(bool loaded) {isImageLoaded = loaded;}


protected:
   static LRESULT CALLBACK WndProc(HWND hWnd, UINT uMessage, WPARAM wParam, LPARAM lParam);
   virtual bool wantMsg(UINT uMsg);
   virtual bool evMsg(HWND hWnd, UINT uMsg, WPARAM wParam,
                         LPARAM lParam);

   // Windows instance handle of this process
   static HINSTANCE _hInstance;

   // Windows window handle of default window
   HWND _hwnd;

   // Reference to object used by JNI functions in this handler
   jobject _handlerObj;

   // Reference to Java object passed by NativeHandler.enable()
   jobject _exposedObj;

   // Pointer to WindowThread instance belonging to this handler
   // Only one thread is created for all handlers in the same Java application
   static WindowThread * _pThread;

   // Window position (ulx, uly)
   int _xPos;
   int _yPos;

   // Window size (width, height)
   int _winWidth;
   int _winHeight;
   
   // Window style   
   DWORD _winStyle;

   // Window title in title bar   
   char * _title;

private:
   enum
   {
      enableCode = 1,
      disableCode = 2,
      suspendCode = 3
   };

   void doEnable();
   void destroyBuf();
   void destroyIconEventBuf();
   void killJNIThread(); // JNIThread
   void iconCallback(int msg);
   bool setIconData();
   void reportIconEvent(int evt);
   void setDetached(bool detached);

   bool isLockedRead;
   bool isLockedWrite;
   jmethodID evtMsgId;
   jfieldID dataBlockId;
   bool isInitDone;
   volatile bool isDetached;
   bool isTransparent;
   JNIBuffer<jbooleanArray> * _pBoolBuf;
   JNIBuffer<jcharArray> * _pCharBuf;
   JNIBuffer<jbyteArray> * _pByteBuf;
   JNIBuffer<jshortArray> * _pShortBuf;
   JNIBuffer<jintArray> * _pIntBuf;
   JNIBuffer<jlongArray> * _pLongBuf;
   JNIBuffer<jfloatArray> * _pFloatBuf;
   JNIBuffer<jdoubleArray> * _pDoubleBuf;

   JNIThread * pJNIThread;
   bool isImageLoaded;
   
   HICON hIcon;
   int iconEventMask;  
   bool isTrayIcon;
   JNIBuffer<jintArray> * _pIconEventBuf;
   NOTIFYICONDATA iconData;

   static DWORD WINAPI ThreadProc(LPVOID lpParameter);
   static DWORD getDllVersion(LPCTSTR lpszDllName);
   static PSLWA pSetLayeredWindowAttributes;
   static bool isKernelInitialized;
};

#endif
